# Copyright (C) 2022 Greenbone AG
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# pylint: disable=redefined-builtin

from unittest.mock import MagicMock

import httpx

from pontos.github.api.labels import GitHubAsyncRESTLabels
from tests import AsyncIteratorMock, aiter, anext
from tests.github.api import GitHubAsyncRESTTestCase, create_response


class GitHubAsyncRESTLabelsTestCase(GitHubAsyncRESTTestCase):
    api_cls = GitHubAsyncRESTLabels

    async def test_get_all(self):
        response1 = create_response()
        response1.json.return_value = [{"id": 1, "name": "a"}]
        response2 = create_response()
        response2.json.return_value = [
            {"id": 2, "name": "b"},
            {"id": 3, "name": "c"},
        ]

        self.client.get_all.return_value = AsyncIteratorMock(
            [response1, response2]
        )

        async_it = aiter(self.api.get_all("foo/bar", 123))
        label = await anext(async_it)
        self.assertEqual(label, "a")
        label = await anext(async_it)
        self.assertEqual(label, "b")
        label = await anext(async_it)
        self.assertEqual(label, "c")

        with self.assertRaises(StopAsyncIteration):
            await anext(async_it)

        self.client.get_all.assert_called_once_with(
            "/repos/foo/bar/issues/123/labels",
            params={"per_page": "100"},
        )

    async def test_set_all(self):
        response = create_response()
        self.client.post.return_value = response

        await self.api.set_all("foo/bar", 123, ["a", "b"])

        self.client.post.assert_awaited_once_with(
            "/repos/foo/bar/issues/123/labels", data={"labels": ["a", "b"]}
        )

    async def test_set_labels_failure(self):
        response = create_response()
        self.client.post.side_effect = httpx.HTTPStatusError(
            "404", request=MagicMock(), response=response
        )

        with self.assertRaises(httpx.HTTPStatusError):
            await self.api.set_all("foo/bar", 123, ["a", "b"])

        self.client.post.assert_awaited_once_with(
            "/repos/foo/bar/issues/123/labels", data={"labels": ["a", "b"]}
        )



# Copyright (C) 2021-2022 Greenbone AG
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import re
from collections import defaultdict
from datetime import date
from pathlib import Path
from typing import NamedTuple, Optional, TypedDict, Union

import tomlkit

from pontos.changelog.errors import ChangelogBuilderError
from pontos.git import Git
from pontos.typing import SupportsStr

ADDRESS = "https://github.com/"

DEFAULT_CHANGELOG_CONFIG = """commit_types = [
    { message = "^add", group = "Added"},
    { message = "^remove", group = "Removed"},
    { message = "^change", group = "Changed"},
    { message = "^fix", group = "Bug Fixes"},
    { message = "^deps", group = "Dependencies"},
]
"""


class CommitType(TypedDict):
    message: str
    group: str


class CommitLogEntry(NamedTuple):
    commit_id: str
    message: str


class ConventionalCommits:
    """
    Extracts conventional commits from the git log

    Example:
        Collect the conventional commits between the tags "v1.2.3" and "v2.0.0"
        using the default config settings. Afterwards get the list of commits
        for the "Added" category.

        .. code-block:: python

            from pontos.changelog import ConventionalCommits

            collector = ConventionalCommits(space="my-org", project="my-project)
            commits = collector.get_commits(
                from_ref="v1.2.3",
                to_ref="v2.0.0",
            )
            added = commits.get("Added")
    """

    def __init__(
        self,
        config: Optional[Path] = None,
    ) -> None:
        """
        Create a new ConventionalCommits instance for collecting conventional
        commits from a git log.

        Args:
            config: Optional TOML config for conventional commit parsing
                settings.
        """
        if config:
            if not config.exists():
                raise ChangelogBuilderError(
                    f"Changelog Config file '{config.absolute()}' does not "
                    "exist."
                )

            self._config = tomlkit.parse(config.read_text(encoding="utf-8"))
        else:
            self._config = tomlkit.parse(DEFAULT_CHANGELOG_CONFIG)

    def get_commits(
        self,
        from_ref: Optional[SupportsStr] = None,
        to_ref: SupportsStr = "HEAD",
    ) -> dict[str, list[CommitLogEntry]]:
        """
        Get all commits by conventional commit type between a range of git
        references.

        Args:
            from_ref: Git commit ID or reference where to start looking for
                conventional commits. If None, to_ref is ignored and all
                 conventional commits are returned.
            to_ref: Git commit ID or reference where to stop looking for
                conventional commits. By default HEAD is used.

        Returns:
            A dict containing the grouped log entries
        """
        commit_list = self._get_git_log(from_ref, to_ref)
        return self._sort_commits(commit_list)

    def commit_types(self) -> list[CommitType]:
        return self._config.get("commit_types", [])

    def _get_git_log(
        self, from_ref: Optional[SupportsStr], to_ref: SupportsStr = "HEAD"
    ) -> list[str]:
        """Getting the git log for the a range of commits.

        Requires the fitting branch to be checked out if to_ref is not set.

        Args:
            from_ref: Git commit ID or reference of the first log entry. If
                None, to_ref is ignored and all log entries of the current
                checked out branch are returned.
            to_ref: Git commit ID or reference where to stop considering the
                log entries. By default HEAD is used which points to the last
                commit of the checked out branch.

        Returns:
            A list of `git log` entries
        """
        git = Git()
        if not from_ref:
            return git.log(oneline=True)

        return git.log(
            f"{from_ref}..{to_ref}",
            oneline=True,
        )

    def _sort_commits(
        self, commits: list[str]
    ) -> dict[str, list[CommitLogEntry]]:
        """Sort the commits by commit type and group them
        in a dict
        ```
        {
            'Added:': [
                ('commit 1', 'message 1'),
                ('commit 2', 'message 2'),
                '...',
            ],
            'Fixed:': [
                ...
            ],
        }
        ```

        Returns
            The dict containing the commit messages
        """
        expressions = [
            (
                commit_type["group"],
                re.compile(rf'{commit_type["message"]}\s?[:|-]', flags=re.I),
            )
            for commit_type in self.commit_types()
        ]
        commit_dict = defaultdict(list)
        if commits and len(commits) > 0:
            for commit in commits:
                commit_id, message = commit.split(" ", maxsplit=1)
                for group, reg in expressions:
                    match = reg.match(message)
                    if match:
                        # remove the commit tag from commit message
                        cleaned_msg = message.replace(
                            match.group(0), ""
                        ).strip()
                        commit_dict[group].append(
                            CommitLogEntry(
                                commit_id=commit_id, message=cleaned_msg
                            )
                        )

        return commit_dict


class ChangelogBuilder:
    """
    Creates Changelog from conventional commits using the git log
    from the latest version.

    Example:
        Create a changelog as a string from the changes between git tags
        "v1.2.3" and "v2.0.0" using the default config settings.

        .. code-block:: python

            from pontos.changelog import ChangelogBuilder

            builder = ChangelogBuilder(space="my-org", project="my-project)
            changelog = builder.create_changelog(
                last_version="1.2.3",
                next_version="2.0.0",
            )
    """

    def __init__(
        self,
        *,
        space: str,
        project: str,
        git_tag_prefix: Optional[str] = "v",
        config: Optional[Path] = None,
    ) -> None:
        """
        Create a new ChangelogBuilder instance.

        Args:
            space: GitHub space to use (organization or user name).
            project: GitHub project to create a changelog for
            git_tag_prefix: Git tag prefix to use when checking for git tags.
                Default is "v".
            config: TOML config for conventional commit parsing settings
        """
        self._project = project
        self._space = space

        self._git_tag_prefix = git_tag_prefix
        self._conventional_commits = ConventionalCommits(config)

    def create_changelog(
        self,
        *,
        last_version: Optional[SupportsStr] = None,
        next_version: Optional[SupportsStr] = None,
    ) -> str:
        """
        Create a changelog

        Args:
            last_version: Version of the last release. If None it is considered
                as the first release.
            next_version: Version of the to be created release the changelog
                corresponds to. If None a changelog for an unrelease version
                will be created.

        Returns:
            The created changelog content.
        """
        commit_dict = self._conventional_commits.get_commits(
            f"{self._git_tag_prefix}{last_version}" if last_version else None
        )
        return self._build_changelog(last_version, next_version, commit_dict)

    def create_changelog_file(
        self,
        output: Union[str, Path],
        *,
        last_version: Optional[SupportsStr] = None,
        next_version: Optional[SupportsStr] = None,
    ) -> None:
        """
        Create a changelog and write the changelog to a file

        Args:
            output: A file path where to store the changelog
            last_version: Version of the last release. If None it is considered
                as the first release.
            next_version: Version of the to be created release the changelog
                corresponds to. If None a changelog for an unrelease version
                will be created.
        """
        changelog = self.create_changelog(
            last_version=last_version, next_version=next_version
        )
        self._write_changelog_file(changelog, output)

    def _get_first_commit(self) -> str:
        """
        Git the first commit ID for the current branch
        """
        git = Git()
        return git.rev_list("HEAD", max_parents=0, abbrev_commit=True)[0]

    def _build_changelog(
        self,
        last_version: Optional[SupportsStr],
        next_version: Optional[SupportsStr],
        commit_dict: dict[str, list[CommitLogEntry]],
    ) -> str:
        """
        Building the changelog from the passed commit information.

        Args:
            commit_dict: dict containing sorted commits

        Returns:
            The changelog content
        """

        # changelog header
        changelog = []
        if next_version:
            changelog.append(
                f"## [{next_version}] - {date.today().isoformat()}"
            )
        else:
            changelog.append("## [Unreleased]")

        # changelog entries
        for commit_type in self._conventional_commits.commit_types():
            if commit_type["group"] in commit_dict.keys():
                changelog.append(f"\n## {commit_type['group']}")
                for log_entry in commit_dict[commit_type["group"]]:
                    commit_id, commit_message = log_entry
                    commit_link = (
                        f"{ADDRESS}{self._space}/{self._project}/"
                        f"commit/{commit_id}"
                    )
                    msg = f"{commit_message} [{commit_id}]({commit_link})"
                    changelog.append(f"* {msg}")

        # comparison line (footer)
        pre = "\n[Unreleased]: "
        compare_link = f"{ADDRESS}{self._space}/{self._project}/compare/"
        if next_version and last_version:
            pre = f"\n[{next_version}]: "
            diff = (
                f"{self._git_tag_prefix}{last_version}..."
                f"{self._git_tag_prefix}{next_version}"
            )
        elif next_version:
            first_commit = self._get_first_commit()
            pre = f"\n[{next_version}]: "
            diff = f"{first_commit}...{self._git_tag_prefix}{next_version}"
        elif last_version:
            # unreleased version
            diff = f"{self._git_tag_prefix}{last_version}...HEAD"
        else:
            # unreleased version
            first_commit = self._get_first_commit()
            diff = f"{first_commit}...HEAD"

        changelog.append(f"{pre}{compare_link}{diff}")

        return "\n".join(changelog)

    def _write_changelog_file(
        self, changelog: str, output: Union[str, Path]
    ) -> None:
        """
        Write changelog to an output file

        Args:
            changelog: Changelog content to write to output file
            output: File name to write changelog into
        """

        changelog_file = Path(output)

        changelog_dir = changelog_file.parent
        changelog_dir.mkdir(parents=True, exist_ok=True)

        changelog_file.write_text(changelog, encoding="utf-8")



# Copyright (C) 2022 Greenbone AG
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# pylint: disable=line-too-long
# ruff: noqa: E501

import unittest
from datetime import date, datetime

from pontos.nvd.models import cvss_v2, cvss_v3
from pontos.nvd.models.cve import CVE, CVSSType, Operator
from tests.nvd import get_cve_data


class CVETestCase(unittest.TestCase):
    def test_required_only(self):
        cve = CVE.from_dict(get_cve_data())

        self.assertEqual(cve.id, "CVE-2022-45536")
        self.assertEqual(cve.source_identifier, "cve@mitre.org")
        self.assertEqual(
            cve.published, datetime(2022, 11, 22, 21, 15, 11, 103000)
        )
        self.assertEqual(
            cve.last_modified, datetime(2022, 11, 23, 16, 2, 7, 367000)
        )
        self.assertEqual(len(cve.descriptions), 1)
        self.assertEqual(len(cve.references), 2)
        self.assertEqual(len(cve.weaknesses), 0)
        self.assertEqual(len(cve.configurations), 0)
        self.assertEqual(len(cve.vendor_comments), 0)
        self.assertIsNone(cve.metrics)
        self.assertIsNone(cve.evaluator_comment)
        self.assertIsNone(cve.evaluator_solution)
        self.assertIsNone(cve.evaluator_impact)
        self.assertIsNone(cve.cisa_exploit_add)
        self.assertIsNone(cve.cisa_action_due)
        self.assertIsNone(cve.cisa_required_action)
        self.assertIsNone(cve.cisa_vulnerability_name)

    def test_descriptions(self):
        cve = CVE.from_dict(get_cve_data())

        description = cve.descriptions[0]
        self.assertEqual(description.lang, "en")
        self.assertEqual(
            description.value,
            "AeroCMS v0.0.1 was discovered to contain a SQL "
            "Injection vulnerability via the id parameter at "
            "\\admin\\post_comments.php. This vulnerability allows "
            "attackers to access database information.",
        )

    def test_references(self):
        cve = CVE.from_dict(get_cve_data())

        reference = cve.references[0]
        self.assertEqual(
            reference.url,
            "https://github.com/rdyx0/CVE/blob/master/AeroCMS/AeroCMS-v0.0.1-SQLi/post_comments_sql_injection/post_comments_sql_injection.md",
        )
        self.assertEqual(reference.source, "cve@mitre.org")
        self.assertEqual(reference.tags, ["Exploit", "Third Party Advisory"])

        reference = cve.references[1]
        self.assertEqual(
            reference.url,
            "https://rdyx0.github.io/2018/09/07/AeroCMS-v0.0.1-SQLi%20post_comments_sql_injection/",
        )
        self.assertEqual(reference.source, "cve@mitre.org")
        self.assertEqual(reference.tags, ["Exploit", "Third Party Advisory"])

    def test_weaknesses(self):
        cve = CVE.from_dict(
            get_cve_data(
                {
                    "weaknesses": [
                        {
                            "source": "nvd@nist.gov",
                            "type": "Primary",
                            "description": [{"lang": "en", "value": "CWE-89"}],
                        }
                    ],
                }
            )
        )

        self.assertEqual(len(cve.weaknesses), 1)

        weakness = cve.weaknesses[0]
        self.assertEqual(weakness.source, "nvd@nist.gov")
        self.assertEqual(weakness.type, "Primary")
        self.assertEqual(len(weakness.description), 1)

        description = weakness.description[0]
        self.assertEqual(description.lang, "en")
        self.assertEqual(description.value, "CWE-89")

    def test_configuration(self):
        cve = CVE.from_dict(
            get_cve_data(
                {
                    "configurations": [
                        {
                            "nodes": [
                                {
                                    "operator": "OR",
                                    "negate": False,
                                    "cpe_match": [
                                        {
                                            "vulnerable": True,
                                            "criteria": "cpe:2.3:a:aerocms_project:aerocms:0.0.1:*:*:*:*:*:*:*",
                                            "match_criteria_id": "52639E84-244D-4DA3-B1AE-6D8BA1C38863",
                                        }
                                    ],
                                }
                            ]
                        }
                    ],
                }
            )
        )

        self.assertEqual(len(cve.configurations), 1)

        configuration = cve.configurations[0]
        self.assertIsNone(configuration.operator)
        self.assertIsNone(configuration.negate)
        self.assertEqual(len(configuration.nodes), 1)

        node = configuration.nodes[0]
        self.assertFalse(node.negate)
        self.assertEqual(node.operator, Operator.OR)
        self.assertEqual(len(node.cpe_match), 1)

        cpe_match = node.cpe_match[0]
        self.assertTrue(cpe_match.vulnerable)
        self.assertEqual(
            cpe_match.criteria,
            "cpe:2.3:a:aerocms_project:aerocms:0.0.1:*:*:*:*:*:*:*",
        )
        self.assertEqual(
            cpe_match.match_criteria_id, "52639E84-244D-4DA3-B1AE-6D8BA1C38863"
        )
        self.assertIsNone(cpe_match.version_start_including)
        self.assertIsNone(cpe_match.version_start_excluding)
        self.assertIsNone(cpe_match.version_end_including)
        self.assertIsNone(cpe_match.version_end_excluding)

    def test_metrics_v2(self):
        cve = CVE.from_dict(
            get_cve_data(
                {
                    "metrics": {
                        "cvss_metric_v2": [
                            {
                                "source": "nvd@nist.gov",
                                "type": "Primary",
                                "cvss_data": {
                                    "version": "2.0",
                                    "vector_string": "AV:N/AC:M/Au:N/C:N/I:P/A:N",
                                    "access_vector": "NETWORK",
                                    "access_complexity": "MEDIUM",
                                    "authentication": "NONE",
                                    "confidentiality_impact": "NONE",
                                    "integrity_impact": "PARTIAL",
                                    "availability_impact": "NONE",
                                    "base_score": 4.3,
                                    "base_severity": "MEDIUM",
                                },
                                "exploitability_score": 8.6,
                                "impact_score": 2.9,
                                "ac_insuf_info": False,
                                "obtain_all_privilege": False,
                                "obtain_user_privilege": False,
                                "obtain_other_privilege": False,
                                "user_interaction_required": True,
                            }
                        ]
                    }
                }
            )
        )

        self.assertEqual(len(cve.metrics.cvss_metric_v2), 1)
        self.assertEqual(len(cve.metrics.cvss_metric_v30), 0)
        self.assertEqual(len(cve.metrics.cvss_metric_v31), 0)

        cvss_metric = cve.metrics.cvss_metric_v2[0]
        self.assertEqual(cvss_metric.source, "nvd@nist.gov")
        self.assertEqual(cvss_metric.type, CVSSType.PRIMARY)
        self.assertEqual(cvss_metric.exploitability_score, 8.6)
        self.assertEqual(cvss_metric.impact_score, 2.9)
        self.assertFalse(cvss_metric.ac_insuf_info)
        self.assertFalse(cvss_metric.obtain_all_privilege)
        self.assertFalse(cvss_metric.obtain_user_privilege)
        self.assertFalse(cvss_metric.obtain_other_privilege)
        self.assertTrue(cvss_metric.user_interaction_required)

        cvss_data = cvss_metric.cvss_data
        self.assertEqual(cvss_data.version, "2.0")
        self.assertEqual(cvss_data.vector_string, "AV:N/AC:M/Au:N/C:N/I:P/A:N")
        self.assertEqual(cvss_data.base_score, 4.3)
        self.assertEqual(cvss_data.access_vector, cvss_v2.AccessVector.NETWORK)
        self.assertEqual(
            cvss_data.access_complexity, cvss_v2.AccessComplexity.MEDIUM
        )
        self.assertEqual(cvss_data.authentication, cvss_v2.Authentication.NONE)
        self.assertEqual(cvss_data.confidentiality_impact, cvss_v2.Impact.NONE)
        self.assertEqual(cvss_data.integrity_impact, cvss_v2.Impact.PARTIAL)
        self.assertEqual(cvss_data.availability_impact, cvss_v2.Impact.NONE)
        self.assertIsNone(cvss_data.exploitability)
        self.assertIsNone(cvss_data.remediation_level)
        self.assertIsNone(cvss_data.report_confidence)
        self.assertIsNone(cvss_data.temporal_score)
        self.assertIsNone(cvss_data.collateral_damage_potential)
        self.assertIsNone(cvss_data.target_distribution)
        self.assertIsNone(cvss_data.confidentiality_requirement)
        self.assertIsNone(cvss_data.integrity_requirement)
        self.assertIsNone(cvss_data.availability_requirement)
        self.assertIsNone(cvss_data.environmental_score)

    def test_metrics_v30(self):
        cve = CVE.from_dict(
            get_cve_data(
                {
                    "metrics": {
                        "cvss_metric_v30": [
                            {
                                "source": "nvd@nist.gov",
                                "type": "Primary",
                                "cvss_data": {
                                    "version": "3.0",
                                    "vector_string": "CVSS:3.0/AV:N/AC:L/PR:N/UI:R/S:C/C:L/I:L/A:N",
                                    "attack_vector": "NETWORK",
                                    "attack_complexity": "LOW",
                                    "privileges_required": "NONE",
                                    "user_interaction": "REQUIRED",
                                    "scope": "CHANGED",
                                    "confidentiality_impact": "LOW",
                                    "integrity_impact": "LOW",
                                    "availability_impact": "NONE",
                                    "base_score": 6.1,
                                    "base_severity": "MEDIUM",
                                },
                                "exploitability_score": 2.8,
                                "impact_score": 2.7,
                            }
                        ],
                    }
                }
            )
        )

        self.assertEqual(len(cve.metrics.cvss_metric_v2), 0)
        self.assertEqual(len(cve.metrics.cvss_metric_v30), 1)
        self.assertEqual(len(cve.metrics.cvss_metric_v31), 0)

        cvss_metric = cve.metrics.cvss_metric_v30[0]
        self.assertEqual(cvss_metric.source, "nvd@nist.gov")
        self.assertEqual(cvss_metric.type, CVSSType.PRIMARY)
        self.assertEqual(cvss_metric.exploitability_score, 2.8)
        self.assertEqual(cvss_metric.impact_score, 2.7)

        cvss_data = cvss_metric.cvss_data
        self.assertEqual(cvss_data.version, "3.0")
        self.assertEqual(
            cvss_data.vector_string,
            "CVSS:3.0/AV:N/AC:L/PR:N/UI:R/S:C/C:L/I:L/A:N",
        )
        self.assertEqual(cvss_data.base_score, 6.1)
        self.assertEqual(cvss_data.base_severity, cvss_v3.Severity.MEDIUM)
        self.assertEqual(cvss_data.attack_vector, cvss_v3.AttackVector.NETWORK)
        self.assertEqual(
            cvss_data.attack_complexity, cvss_v3.AttackComplexity.LOW
        )
        self.assertEqual(
            cvss_data.privileges_required, cvss_v3.PrivilegesRequired.NONE
        )
        self.assertEqual(
            cvss_data.user_interaction, cvss_v3.UserInteraction.REQUIRED
        )
        self.assertEqual(cvss_data.scope, cvss_v3.Scope.CHANGED)
        self.assertEqual(cvss_data.confidentiality_impact, cvss_v3.Impact.LOW)
        self.assertEqual(cvss_data.integrity_impact, cvss_v3.Impact.LOW)
        self.assertEqual(cvss_data.availability_impact, cvss_v3.Impact.NONE)
        self.assertIsNone(cvss_data.exploit_code_maturity)
        self.assertIsNone(cvss_data.remediation_level)
        self.assertIsNone(cvss_data.report_confidence)
        self.assertIsNone(cvss_data.temporal_score)
        self.assertIsNone(cvss_data.temporal_severity)
        self.assertIsNone(cvss_data.confidentiality_requirement)
        self.assertIsNone(cvss_data.integrity_requirement)
        self.assertIsNone(cvss_data.availability_requirement)
        self.assertIsNone(cvss_data.modified_attack_vector)
        self.assertIsNone(cvss_data.modified_attack_complexity)
        self.assertIsNone(cvss_data.modified_privileges_required)
        self.assertIsNone(cvss_data.modified_user_interaction)
        self.assertIsNone(cvss_data.modified_scope)
        self.assertIsNone(cvss_data.modified_confidentiality_impact)
        self.assertIsNone(cvss_data.modified_integrity_impact)
        self.assertIsNone(cvss_data.modified_availability_impact)
        self.assertIsNone(cvss_data.environmental_score)
        self.assertIsNone(cvss_data.environmental_severity)

    def test_metrics_v31(self):
        cve = CVE.from_dict(
            get_cve_data(
                {
                    "metrics": {
                        "cvss_metric_v31": [
                            {
                                "source": "nvd@nist.gov",
                                "type": "Primary",
                                "cvss_data": {
                                    "version": "3.1",
                                    "vector_string": "CVSS:3.1/AV:N/AC:L/PR:H/UI:N/S:U/C:H/I:N/A:N",
                                    "attack_vector": "NETWORK",
                                    "attack_complexity": "LOW",
                                    "privileges_required": "HIGH",
                                    "user_interaction": "NONE",
                                    "scope": "UNCHANGED",
                                    "confidentiality_impact": "HIGH",
                                    "integrity_impact": "NONE",
                                    "availability_impact": "NONE",
                                    "base_score": 4.9,
                                    "base_severity": "MEDIUM",
                                },
                                "exploitability_score": 1.2,
                                "impact_score": 3.6,
                            }
                        ]
                    },
                }
            )
        )

        self.assertEqual(len(cve.metrics.cvss_metric_v2), 0)
        self.assertEqual(len(cve.metrics.cvss_metric_v30), 0)
        self.assertEqual(len(cve.metrics.cvss_metric_v31), 1)

        cvss_metric = cve.metrics.cvss_metric_v31[0]
        self.assertEqual(cvss_metric.source, "nvd@nist.gov")
        self.assertEqual(cvss_metric.type, CVSSType.PRIMARY)
        self.assertEqual(cvss_metric.exploitability_score, 1.2)
        self.assertEqual(cvss_metric.impact_score, 3.6)

        cvss_data = cvss_metric.cvss_data
        self.assertEqual(cvss_data.version, "3.1")
        self.assertEqual(
            cvss_data.vector_string,
            "CVSS:3.1/AV:N/AC:L/PR:H/UI:N/S:U/C:H/I:N/A:N",
        )
        self.assertEqual(cvss_data.base_score, 4.9)
        self.assertEqual(cvss_data.base_severity, cvss_v3.Severity.MEDIUM)
        self.assertEqual(cvss_data.attack_vector, cvss_v3.AttackVector.NETWORK)
        self.assertEqual(
            cvss_data.attack_complexity, cvss_v3.AttackComplexity.LOW
        )
        self.assertEqual(
            cvss_data.privileges_required, cvss_v3.PrivilegesRequired.HIGH
        )
        self.assertEqual(
            cvss_data.user_interaction, cvss_v3.UserInteraction.NONE
        )
        self.assertEqual(cvss_data.scope, cvss_v3.Scope.UNCHANGED)
        self.assertEqual(cvss_data.confidentiality_impact, cvss_v3.Impact.HIGH)
        self.assertEqual(cvss_data.integrity_impact, cvss_v3.Impact.NONE)
        self.assertEqual(cvss_data.availability_impact, cvss_v3.Impact.NONE)
        self.assertIsNone(cvss_data.exploit_code_maturity)
        self.assertIsNone(cvss_data.remediation_level)
        self.assertIsNone(cvss_data.report_confidence)
        self.assertIsNone(cvss_data.temporal_score)
        self.assertIsNone(cvss_data.temporal_severity)
        self.assertIsNone(cvss_data.confidentiality_requirement)
        self.assertIsNone(cvss_data.integrity_requirement)
        self.assertIsNone(cvss_data.availability_requirement)
        self.assertIsNone(cvss_data.modified_attack_vector)
        self.assertIsNone(cvss_data.modified_attack_complexity)
        self.assertIsNone(cvss_data.modified_privileges_required)
        self.assertIsNone(cvss_data.modified_user_interaction)
        self.assertIsNone(cvss_data.modified_scope)
        self.assertIsNone(cvss_data.modified_confidentiality_impact)
        self.assertIsNone(cvss_data.modified_integrity_impact)
        self.assertIsNone(cvss_data.modified_availability_impact)
        self.assertIsNone(cvss_data.environmental_score)
        self.assertIsNone(cvss_data.environmental_severity)

    def test_metrics_v31_severity_none(self):
        cve = CVE.from_dict(
            get_cve_data(
                {
                    "metrics": {
                        "cvss_metric_v31": [
                            {
                                "source": "nvd@nist.gov",
                                "type": "Secondary",
                                "cvss_data": {
                                    "version": "3.1",
                                    "vector_string": "CVSS:3.1/AV:N/AC:L/PR:H/UI:R/S:U/C:N/I:N/A:N",
                                    "attack_vector": "NETWORK",
                                    "attack_complexity": "LOW",
                                    "privileges_required": "NONE",
                                    "user_interaction": "REQUIRED",
                                    "scope": "UNCHANGED",
                                    "confidentiality_impact": "NONE",
                                    "integrity_impact": "NONE",
                                    "availability_impact": "NONE",
                                    "base_score": 0.0,
                                    "base_severity": "NONE",
                                },
                                "exploitability_score": 2.8,
                                "impact_score": 0.0,
                            }
                        ]
                    },
                }
            )
        )

        self.assertEqual(len(cve.metrics.cvss_metric_v2), 0)
        self.assertEqual(len(cve.metrics.cvss_metric_v30), 0)
        self.assertEqual(len(cve.metrics.cvss_metric_v31), 1)

        cvss_metric = cve.metrics.cvss_metric_v31[0]
        self.assertEqual(cvss_metric.source, "nvd@nist.gov")
        self.assertEqual(cvss_metric.type, CVSSType.SECONDARY)
        self.assertEqual(cvss_metric.exploitability_score, 2.8)
        self.assertEqual(cvss_metric.impact_score, 0.0)

        cvss_data = cvss_metric.cvss_data
        self.assertEqual(cvss_data.version, "3.1")
        self.assertEqual(
            cvss_data.vector_string,
            "CVSS:3.1/AV:N/AC:L/PR:H/UI:R/S:U/C:N/I:N/A:N",
        )
        self.assertEqual(cvss_data.base_score, 0.0)
        self.assertEqual(cvss_data.base_severity, cvss_v3.Severity.NONE)
        self.assertEqual(cvss_data.attack_vector, cvss_v3.AttackVector.NETWORK)
        self.assertEqual(
            cvss_data.attack_complexity, cvss_v3.AttackComplexity.LOW
        )
        self.assertEqual(
            cvss_data.privileges_required, cvss_v3.PrivilegesRequired.NONE
        )
        self.assertEqual(
            cvss_data.user_interaction, cvss_v3.UserInteraction.REQUIRED
        )
        self.assertEqual(cvss_data.scope, cvss_v3.Scope.UNCHANGED)
        self.assertEqual(cvss_data.confidentiality_impact, cvss_v3.Impact.NONE)
        self.assertEqual(cvss_data.integrity_impact, cvss_v3.Impact.NONE)
        self.assertEqual(cvss_data.availability_impact, cvss_v3.Impact.NONE)
        self.assertIsNone(cvss_data.exploit_code_maturity)
        self.assertIsNone(cvss_data.remediation_level)
        self.assertIsNone(cvss_data.report_confidence)
        self.assertIsNone(cvss_data.temporal_score)
        self.assertIsNone(cvss_data.temporal_severity)
        self.assertIsNone(cvss_data.confidentiality_requirement)
        self.assertIsNone(cvss_data.integrity_requirement)
        self.assertIsNone(cvss_data.availability_requirement)
        self.assertIsNone(cvss_data.modified_attack_vector)
        self.assertIsNone(cvss_data.modified_attack_complexity)
        self.assertIsNone(cvss_data.modified_privileges_required)
        self.assertIsNone(cvss_data.modified_user_interaction)
        self.assertIsNone(cvss_data.modified_scope)
        self.assertIsNone(cvss_data.modified_confidentiality_impact)
        self.assertIsNone(cvss_data.modified_integrity_impact)
        self.assertIsNone(cvss_data.modified_availability_impact)
        self.assertIsNone(cvss_data.environmental_score)
        self.assertIsNone(cvss_data.environmental_severity)

    def test_vendor_comments(self):
        cve = CVE.from_dict(
            get_cve_data(
                {
                    "vendor_comments": [
                        {
                            "organization": "Apache",
                            "comment": "Fixed in Apache HTTP Server 1.3.12:\nhttp://httpd.apache.org/security/vulnerabilities_13.html",
                            "last_modified": "2008-07-02T00:00:00",
                        }
                    ],
                }
            )
        )

        self.assertEqual(len(cve.vendor_comments), 1)

        comment = cve.vendor_comments[0]
        self.assertEqual(comment.organization, "Apache")
        self.assertEqual(
            comment.comment,
            "Fixed in Apache HTTP Server 1.3.12:\nhttp://httpd.apache.org/security/vulnerabilities_13.html",
        )
        self.assertEqual(comment.last_modified, datetime(2008, 7, 2))

    def test_evaluator_comment(self):
        cve = CVE.from_dict(
            get_cve_data(
                {
                    "evaluator_comment": "Please see the following link for more information:\r\n\r\nhttp://seclists.org/bugtraq/1999/Jan/0215.html"
                }
            )
        )

        self.assertEqual(
            cve.evaluator_comment,
            "Please see the following link for more information:\r\n\r\nhttp://seclists.org/bugtraq/1999/Jan/0215.html",
        )

    def test_evaluator_solution(self):
        cve = CVE.from_dict(
            get_cve_data(
                {
                    "evaluator_solution": "This problem was fixed in Linux kernel 2.2.4 and later releases."
                }
            )
        )

        self.assertEqual(
            cve.evaluator_solution,
            "This problem was fixed in Linux kernel 2.2.4 and later releases.",
        )

    def test_evaluator_impact(self):
        cve = CVE.from_dict(
            get_cve_data(
                {
                    "evaluator_impact": "This Common Vulnerabilities and Exposures (CVE) entry is a configuration issue and not a software flaw. As such, it doesn’t fit in the CVE software flaw list. The Common Vulnerability Scoring System (CVSS) base score for this CVE entry has been set to 0 because this CVE entry has no impact as a software flaw according to CVSS. This does not mean that the configuration issue is not important and there may be security implications relative to computers having this configuration."
                }
            )
        )

        self.assertEqual(
            cve.evaluator_impact,
            "This Common Vulnerabilities and Exposures (CVE) entry is a configuration issue and not a software flaw. As such, it doesn’t fit in the CVE software flaw list. The Common Vulnerability Scoring System (CVSS) base score for this CVE entry has been set to 0 because this CVE entry has no impact as a software flaw according to CVSS. This does not mean that the configuration issue is not important and there may be security implications relative to computers having this configuration.",
        )

    def test_cisa(self):
        cve = CVE.from_dict(
            get_cve_data(
                {
                    "cisa_exploit_add": "2022-03-03",
                    "cisa_action_due": "2022-03-24",
                    "cisa_required_action": "Apply updates per vendor instructions.",
                    "cisa_vulnerability_name": "Microsoft Windows Privilege Escalation Vulnerability",
                }
            )
        )

        self.assertEqual(cve.cisa_action_due, date(2022, 3, 24))
        self.assertEqual(cve.cisa_exploit_add, date(2022, 3, 3))
        self.assertEqual(
            cve.cisa_required_action, "Apply updates per vendor instructions."
        )
        self.assertEqual(
            cve.cisa_vulnerability_name,
            "Microsoft Windows Privilege Escalation Vulnerability",
        )



# Copyright (C) 2022-2023 Greenbone AG
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import json
import unittest
from unittest.mock import MagicMock, patch

from pontos.testing import temp_directory, temp_file
from pontos.version import VersionError
from pontos.version.commands import JavaScriptVersionCommand
from pontos.version.schemes import SemanticVersioningScheme


class GetCurrentJavaScriptVersionCommandTestCase(unittest.TestCase):
    def test_get_current_version(self):
        content = '{"name": "foo", "version": "1.2.3"}'
        with temp_file(content, name="package.json", change_into=True):
            cmd = JavaScriptVersionCommand(SemanticVersioningScheme)
            version = cmd.get_current_version()

            self.assertEqual(
                version, SemanticVersioningScheme.parse_version("1.2.3")
            )

    def test_no_project_file(self):
        with temp_directory(change_into=True), self.assertRaisesRegex(
            VersionError, ".* file not found."
        ):
            cmd = JavaScriptVersionCommand(SemanticVersioningScheme)
            cmd.get_current_version()

    def test_no_package_version(self):
        content = '{"name": "foo"}'
        with temp_file(
            content, name="package.json", change_into=True
        ), self.assertRaisesRegex(VersionError, "Version field missing in"):
            cmd = JavaScriptVersionCommand(SemanticVersioningScheme)
            cmd.get_current_version()

    def test_no_valid_json_in_package_version(self):
        content = "{"
        with temp_file(
            content, name="package.json", change_into=True
        ), self.assertRaisesRegex(VersionError, "No valid JSON found."):
            cmd = JavaScriptVersionCommand(SemanticVersioningScheme)
            cmd.get_current_version()


class UpdateJavaScriptVersionCommandTestCase(unittest.TestCase):
    def test_update_version_file(self):
        content = '{"name":"foo", "version":"1.2.3"}'

        with temp_file(content, name="package.json", change_into=True) as temp:
            cmd = JavaScriptVersionCommand(SemanticVersioningScheme)
            cmd.get_current_version()
            updated = cmd.update_version(
                SemanticVersioningScheme.parse_version("22.4.0")
            )

            self.assertEqual(
                updated.previous,
                SemanticVersioningScheme.parse_version("1.2.3"),
            )
            self.assertEqual(
                updated.new, SemanticVersioningScheme.parse_version("22.4.0")
            )
            self.assertEqual(updated.changed_files, [temp.resolve()])

            with temp.open(mode="r", encoding="utf-8") as fp:
                fake_package = json.load(fp)

            self.assertEqual(fake_package["version"], "22.4.0")

    def test_update_js_version_file(self):
        content = '{"name":"foo", "version":"1.2.3"}'
        js_content = """const foo = "bar";
const VERSION = "1.2.3";
const func = () => ();
"""

        with temp_directory(change_into=True) as temp_dir:
            package_json = temp_dir / "package.json"
            package_json.write_text(content, encoding="utf8")
            js_version_file = (
                temp_dir / JavaScriptVersionCommand.version_file_paths[0]
            )
            js_version_file.parent.mkdir()
            js_version_file.write_text(js_content, encoding="utf8")

            cmd = JavaScriptVersionCommand(SemanticVersioningScheme)
            updated = cmd.update_version(
                SemanticVersioningScheme.parse_version("22.4.0")
            )

            self.assertEqual(
                updated.previous,
                SemanticVersioningScheme.parse_version("1.2.3"),
            )
            self.assertEqual(
                updated.new, SemanticVersioningScheme.parse_version("22.4.0")
            )
            self.assertEqual(
                updated.changed_files,
                [
                    package_json.resolve(),
                    JavaScriptVersionCommand.version_file_paths[0],
                ],
            )

            with package_json.open(mode="r", encoding="utf-8") as fp:
                fake_package = json.load(fp)

            self.assertEqual(fake_package["version"], "22.4.0")

            self.assertEqual(
                js_version_file.read_text(encoding="utf8"),
                'const foo = "bar";\nconst VERSION = "22.4.0";\n'
                "const func = () => ();\n",
            )

    def test_update_js_version_file_with_single_quotes(self):
        content = '{"name":"foo", "version":"1.2.3"}'
        js_content = """const foo = "bar";
const VERSION = '1.2.3';
const func = () => ();
"""

        with temp_directory(change_into=True) as temp_dir:
            package_json = temp_dir / "package.json"
            package_json.write_text(content, encoding="utf8")
            js_version_file = (
                temp_dir / JavaScriptVersionCommand.version_file_paths[0]
            )
            js_version_file.parent.mkdir()
            js_version_file.write_text(js_content, encoding="utf8")

            cmd = JavaScriptVersionCommand(SemanticVersioningScheme)
            updated = cmd.update_version(
                SemanticVersioningScheme.parse_version("22.4.0")
            )

            self.assertEqual(
                updated.previous,
                SemanticVersioningScheme.parse_version("1.2.3"),
            )
            self.assertEqual(
                updated.new, SemanticVersioningScheme.parse_version("22.4.0")
            )
            self.assertEqual(
                updated.changed_files,
                [
                    package_json.resolve(),
                    JavaScriptVersionCommand.version_file_paths[0],
                ],
            )

            with package_json.open(mode="r", encoding="utf-8") as fp:
                fake_package = json.load(fp)

            self.assertEqual(fake_package["version"], "22.4.0")

            self.assertEqual(
                js_version_file.read_text(encoding="utf8"),
                "const foo = \"bar\";\nconst VERSION = '22.4.0';\n"
                "const func = () => ();\n",
            )

    def test_update_version_files(self):
        content = '{"name":"foo", "version":"1.2.3"}'
        file_content = """const foo = "bar";
const VERSION = "1.2.3";
const func = () => ();
"""

        with temp_directory(change_into=True) as temp_dir:
            package_json = temp_dir / "package.json"
            package_json.write_text(content, encoding="utf8")

            js_version_file = (
                temp_dir / JavaScriptVersionCommand.version_file_paths[0]
            )
            js_version_file.parent.mkdir()
            js_version_file.write_text(file_content, encoding="utf8")

            ts_version_file = (
                temp_dir / JavaScriptVersionCommand.version_file_paths[1]
            )
            ts_version_file.write_text(file_content, encoding="utf8")

            cmd = JavaScriptVersionCommand(SemanticVersioningScheme)
            updated = cmd.update_version(
                SemanticVersioningScheme.parse_version("22.4.0")
            )

            self.assertEqual(
                updated.previous,
                SemanticVersioningScheme.parse_version("1.2.3"),
            )
            self.assertEqual(
                updated.new, SemanticVersioningScheme.parse_version("22.4.0")
            )
            self.assertEqual(
                updated.changed_files,
                [
                    package_json.resolve(),
                    JavaScriptVersionCommand.version_file_paths[0],
                    JavaScriptVersionCommand.version_file_paths[1],
                ],
            )

            with package_json.open(mode="r", encoding="utf-8") as fp:
                fake_package = json.load(fp)

            self.assertEqual(fake_package["version"], "22.4.0")

            self.assertEqual(
                js_version_file.read_text(encoding="utf8"),
                'const foo = "bar";\nconst VERSION = "22.4.0";\n'
                "const func = () => ();\n",
            )

    def test_update_version_develop(self):
        content = '{"name":"foo", "version":"1.2.3"}'

        with temp_file(content, name="package.json", change_into=True) as temp:
            cmd = JavaScriptVersionCommand(SemanticVersioningScheme)
            new_version = SemanticVersioningScheme.parse_version("22.4.0-dev1")
            updated = cmd.update_version(new_version)

            self.assertEqual(
                updated.previous,
                SemanticVersioningScheme.parse_version("1.2.3"),
            )
            self.assertEqual(updated.new, new_version)
            self.assertEqual(updated.changed_files, [temp.resolve()])

            with temp.open(mode="r", encoding="utf-8") as fp:
                fake_package = json.load(fp)

            self.assertEqual(fake_package["version"], "22.4.0-dev1")

    def test_no_update(self):
        content = '{"name":"foo", "version":"1.2.3"}'

        with temp_file(content, name="package.json", change_into=True) as temp:
            cmd = JavaScriptVersionCommand(SemanticVersioningScheme)
            updated = cmd.update_version(
                SemanticVersioningScheme.parse_version("1.2.3")
            )

            self.assertEqual(
                updated.previous,
                SemanticVersioningScheme.parse_version("1.2.3"),
            )
            self.assertEqual(
                updated.new, SemanticVersioningScheme.parse_version("1.2.3")
            )
            self.assertEqual(updated.changed_files, [])

            with temp.open(mode="r", encoding="utf-8") as fp:
                fake_package = json.load(fp)

            self.assertEqual(fake_package["version"], "1.2.3")

    def test_forced_update(self):
        content = '{"name":"foo", "version":"1.2.3"}'

        with temp_file(content, name="package.json", change_into=True) as temp:
            cmd = JavaScriptVersionCommand(SemanticVersioningScheme)
            updated = cmd.update_version(
                SemanticVersioningScheme.parse_version("1.2.3"), force=True
            )

            self.assertEqual(
                updated.previous,
                SemanticVersioningScheme.parse_version("1.2.3"),
            )
            self.assertEqual(
                updated.new, SemanticVersioningScheme.parse_version("1.2.3")
            )
            self.assertEqual(updated.changed_files, [temp.resolve()])

            with temp.open(mode="r", encoding="utf-8") as fp:
                fake_package = json.load(fp)

            self.assertEqual(fake_package["version"], "1.2.3")


class VerifyJavaScriptVersionCommandTestCase(unittest.TestCase):
    def test_versions_not_equal(self):
        with patch.object(
            JavaScriptVersionCommand,
            "get_current_version",
            MagicMock(
                return_value=SemanticVersioningScheme.parse_version("1.2.3")
            ),
        ), self.assertRaisesRegex(
            VersionError,
            "Provided version .* does not match the current version .*",
        ):
            cmd = JavaScriptVersionCommand(SemanticVersioningScheme)
            cmd.verify_version(SemanticVersioningScheme.parse_version("22.4.0"))

    def test_verify_success(self):
        with patch.object(
            JavaScriptVersionCommand,
            "get_current_version",
            MagicMock(
                return_value=SemanticVersioningScheme.parse_version("22.4.0")
            ),
        ):
            cmd = JavaScriptVersionCommand(SemanticVersioningScheme)
            cmd.verify_version(SemanticVersioningScheme.parse_version("22.4.0"))

    def test_verify_js_mismatch(self):
        with patch.object(
            JavaScriptVersionCommand,
            "get_current_version",
            MagicMock(
                return_value=SemanticVersioningScheme.parse_version("22.4.0")
            ),
        ), patch.object(
            JavaScriptVersionCommand,
            "_get_current_file_version",
            MagicMock(
                return_value=SemanticVersioningScheme.parse_version("22.5.0")
            ),
        ), self.assertRaisesRegex(
            VersionError,
            "Provided version 22.4.0 does not match the current version 22.5.0 "
            "in src/version.js.",
        ):
            cmd = JavaScriptVersionCommand(SemanticVersioningScheme)
            cmd.verify_version(SemanticVersioningScheme.parse_version("22.4.0"))

    def test_verify_ts_mismatch(self):
        with patch.object(
            JavaScriptVersionCommand,
            "get_current_version",
            MagicMock(
                return_value=SemanticVersioningScheme.parse_version("22.4.0")
            ),
        ), patch.object(
            JavaScriptVersionCommand,
            "_get_current_file_version",
            MagicMock(
                side_effect=[
                    SemanticVersioningScheme.parse_version("22.4.0"),
                    SemanticVersioningScheme.parse_version("22.5.0"),
                ]
            ),
        ), self.assertRaisesRegex(
            VersionError,
            "Provided version 22.4.0 does not match the current version 22.5.0 "
            "in src/version.ts.",
        ):
            cmd = JavaScriptVersionCommand(SemanticVersioningScheme)
            cmd.verify_version(SemanticVersioningScheme.parse_version("22.4.0"))

    def test_verify_current(self):
        with patch.object(
            JavaScriptVersionCommand,
            "get_current_version",
            MagicMock(
                return_value=SemanticVersioningScheme.parse_version("22.4.0")
            ),
        ):
            cmd = JavaScriptVersionCommand(SemanticVersioningScheme)
            cmd.verify_version("current")
            cmd.verify_version(None)

    def test_verify_current_failure(self):
        with temp_directory(change_into=True):
            cmd = JavaScriptVersionCommand(SemanticVersioningScheme)

            with self.assertRaisesRegex(
                VersionError, "^.*package.json file not found"
            ):
                cmd.verify_version("current")

            with self.assertRaisesRegex(
                VersionError, "^.*package.json file not found"
            ):
                cmd.verify_version(None)

    def test_verify_current_js_version_matches(self):
        content = '{"name":"foo", "version":"1.2.3"}'
        js_content = 'const VERSION = "1.2.3";'

        with temp_directory(change_into=True) as temp_dir:
            package_json = temp_dir / "package.json"
            package_json.write_text(content, encoding="utf8")
            js_version_file = (
                temp_dir / JavaScriptVersionCommand.version_file_paths[0]
            )
            js_version_file.parent.mkdir()
            js_version_file.write_text(js_content, encoding="utf8")

            cmd = JavaScriptVersionCommand(SemanticVersioningScheme)
            cmd.verify_version("current")
            cmd.verify_version(None)

    def test_verify_current_js_mismatch(self):
        content = '{"name":"foo", "version":"1.2.3"}'
        js_content = 'const VERSION = "1.2.4";'

        with temp_directory(
            change_into=True
        ) as temp_dir, self.assertRaisesRegex(
            VersionError,
            "The version 1.2.4 in src/version.js doesn't match the current "
            "version 1.2.3.",
        ):
            package_json = temp_dir / "package.json"
            package_json.write_text(content, encoding="utf8")
            js_version_file = (
                temp_dir / JavaScriptVersionCommand.version_file_paths[0]
            )
            js_version_file.parent.mkdir()
            js_version_file.write_text(js_content, encoding="utf8")

            cmd = JavaScriptVersionCommand(SemanticVersioningScheme)
            cmd.verify_version("current")
            cmd.verify_version(None)


class ProjectFileJavaScriptVersionCommandTestCase(unittest.TestCase):
    def test_project_file_not_found(self):
        with temp_directory(change_into=True):
            cmd = JavaScriptVersionCommand(SemanticVersioningScheme)

            self.assertFalse(cmd.project_found())

    def test_project_file_found(self):
        with temp_file(name="package.json", change_into=True):
            cmd = JavaScriptVersionCommand(SemanticVersioningScheme)

            self.assertTrue(cmd.project_found())



# Copyright (C) 2022 Greenbone AG
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import asyncio
from pathlib import Path
from typing import (
    AsyncContextManager,
    AsyncIterator,
    Iterable,
    Optional,
    Tuple,
    Union,
)

import httpx

from pontos.github.api.client import GitHubAsyncREST
from pontos.github.api.helper import JSON_OBJECT
from pontos.github.models.release import Release
from pontos.helper import AsyncDownloadProgressIterable, download_async, upload


class GitHubAsyncRESTReleases(GitHubAsyncREST):
    async def create(
        self,
        repo: str,
        tag: str,
        *,
        body: Optional[str] = None,
        name: Optional[str] = None,
        target_commitish: Optional[str] = None,
        draft: bool = False,
        prerelease: bool = False,
    ) -> Release:
        """
        Create a new GitHub release

        https://docs.github.com/en/rest/releases/releases#create-a-release

        Args:
            repo: GitHub repository (owner/name) to use
            tag: The git tag for the release
            body: Content of the changelog for the release
            name: name of the release, e.g. 'pontos 1.0.0'
            target_commitish: Only needed when tag is not there yet
            draft: If the release is a draft. False by default.
            prerelease: If the release is a pre release. False by default.

        Raises:
            httpx.HTTPStatusError: If the request was invalid

        Example:
            .. code-block:: python

                from pontos.github.api import GitHubAsyncRESTApi

                async with GitHubAsyncRESTApi(token) as api:
                    await api.releases.create(
                        "foo/bar",
                        "v1.2.3",
                        body="A new release",
                        name="Bar Release 1.2.3",
                    )
        """
        data: JSON_OBJECT = {
            "tag_name": tag,
            "draft": draft,
            "prerelease": prerelease,
        }
        if name is not None:
            data["name"] = name
        if body is not None:
            data["body"] = body
        if target_commitish is not None:
            data["target_commitish"] = target_commitish

        api = f"/repos/{repo}/releases"
        response = await self._client.post(api, data=data)
        response.raise_for_status()
        return Release.from_dict(response.json())

    async def exists(self, repo: str, tag: str) -> bool:
        """
        Check wether a GitHub release exists by tag

        Args:
            repo: GitHub repository (owner/name) to use
            tag: The git tag for the release

        Returns:
            True if the release exists

        Example:
            .. code-block:: python

                from pontos.github.api import GitHubAsyncRESTApi

                async with GitHubAsyncRESTApi(token) as api:
                    exists = await = api.releases.exists("foo/bar", "v1.2.3")
        """
        api = f"/repos/{repo}/releases/tags/{tag}"
        response = await self._client.get(api)
        return response.is_success

    async def get(self, repo: str, tag: str) -> Release:
        """
        Get data of a GitHub release by tag

        https://docs.github.com/en/rest/releases/releases#get-a-release-by-tag-name

        Args:
            repo: GitHub repository (owner/name) to use
            tag: The git tag for the release

        Raises:
            httpx.HTTPStatusError: If the request was invalid

        Returns:
            Information about the release

        Example:
            .. code-block:: python

                from pontos.github.api import GitHubAsyncRESTApi

                async with GitHubAsyncRESTApi(token) as api:
                    release = await api.releases.get("foo/bar", "v1.2.3)
                    print(release)
        """
        api = f"/repos/{repo}/releases/tags/{tag}"
        response = await self._client.get(api)
        response.raise_for_status()
        return Release.from_dict(response.json())

    def download_release_tarball(
        self, repo: str, tag: str
    ) -> AsyncContextManager[AsyncDownloadProgressIterable[bytes]]:
        # pylint: disable=line-too-long
        """
        Download a release tarball (tar.gz) file

        Args:
            repo: GitHub repository (owner/name) to use
            tag: The git tag for the release

        Raises:
            HTTPStatusError: If the request was invalid

        Example:
            .. code-block:: python

                from pontos.github.api import GitHubAsyncRESTApi

                async with GitHubAsyncRESTApi(token) as api, async with api.releases.download_release_tarball(
                    "foo/bar", "v1.0.0"
                ) as download:
                    file_path = Path("a.file.tar.gz")
                    with file_path.open("wb") as f:
                        async for content, progress in download:
                            f.write(content)
                            print(progress)
        """  # noqa: E501
        api = f"https://github.com/{repo}/archive/refs/tags/{tag}.tar.gz"
        return download_async(self._client.stream(api), url=api)

    def download_release_zip(
        self,
        repo: str,
        tag: str,
    ) -> AsyncContextManager[AsyncDownloadProgressIterable[bytes]]:
        # pylint: disable=line-too-long
        """
        Download a release zip file

        Args:
            repo: GitHub repository (owner/name) to use
            tag: The git tag for the release

        Raises:
            HTTPStatusError: If the request was invalid

        Example:
            .. code-block:: python

                from pontos.github.api import GitHubAsyncRESTApi

                async with GitHubAsyncRESTApi(token) as api, async with api.releases.download_release_zip(
                    "foo/bar", "v1.0.0"
                ) as download:
                    file_path = Path("a.file.zip")
                    with file_path.open("wb") as f:
                        async for content, progress in download:
                            f.write(content)
                            print(progress)
        """  # noqa: E501
        api = f"https://github.com/{repo}/archive/refs/tags/{tag}.zip"
        return download_async(self._client.stream(api), url=api)

    async def download_release_assets(
        self,
        repo: str,
        tag: str,
        *,
        match_pattern: Optional[str] = None,
    ) -> AsyncIterator[
        Tuple[str, AsyncContextManager[AsyncDownloadProgressIterable[bytes]]]
    ]:
        # pylint: disable=line-too-long
        """
        Download release assets

        Args:
            repo: GitHub repository (owner/name) to use
            tag: The git tag for the release
            match_pattern: Optional pattern which the name of the available
                artifact must match. For example "*.zip". Allows to download
                only specific artifacts.

        Raises:
            HTTPStatusError: If the request was invalid

        Example:
            .. code-block:: python

                from pontos.github.api import GitHubAsyncRESTApi

                async with GitHubAsyncRESTApi(token) as api:
                    async def download_asset(name: str, download_cm) -> Path:
                        async with download_cm as iterator:
                            file_path = Path(name)
                            with file_path.open("wb") as f:
                                async for content, progress in iterator:
                                    f.write(content)
                                    print(name, progress)
                            return file_path


                    tasks = []
                    async for name, download_cm in api.releases.download_release_assets(
                        "foo/bar, "v1.2.3",
                    ):
                        tasks.append(asyncio.create_task(
                            download_asset(name, download_cm)
                        )

                    file_paths = await asyncio.gather(*tasks)
        """  # noqa: E501
        release = await self.get(repo, tag)
        assets_url = release.assets_url
        if not assets_url:
            raise RuntimeError("assets URL not found")

        response = await self._client.get(assets_url)
        response.raise_for_status()

        assets_json = response.json()
        for asset_json in assets_json:
            # use browser_download_url here because url doesn't response with
            # exactly the same data on every request.
            # not getting exactly the same data changes the hash sum.
            asset_url: str = asset_json.get("browser_download_url", "")
            name: str = asset_json.get("name", "")

            if match_pattern and not Path(name).match(match_pattern):
                continue

            yield name, download_async(
                self._client.stream(asset_url), url=asset_url
            )

    async def upload_release_assets(
        self,
        repo: str,
        tag: str,
        files: Iterable[Union[Path, Tuple[Path, str]]],
    ) -> AsyncIterator[Path]:
        # pylint: disable=line-too-long
        """
        Upload release assets asynchronously

        Args:
            repo: GitHub repository (owner/name) to use
            tag: The git tag for the release
            files: An iterable of file paths or an iterable of tuples
                containing a file path and content types to upload as an asset

        Returns:
            yields each file after its upload is finished

        Raises:
            HTTPStatusError: If an upload request was invalid

        Example:
            .. code-block:: python

                from pontos.github.api import GitHubAsyncRESTApi

                async with GitHubAsyncRESTApi(token) as api:

                    files = (Path("foo.txt"), Path("bar.txt"),)
                    async for uploaded_file in api.releases.upload_release_assets(
                        "foo/bar", "1.2.3", files
                    ):
                        print(f"Uploaded: {uploaded_file}")

                    files = [
                        (Path("foo.txt"), "text/ascii"),
                        (Path("bar.pdf"), "application/pdf"),
                    ]
                    async for uploaded_file in api.releases.upload_release_assets(
                        "foo/bar", "1.2.3", files
                    ):
                        print(f"Uploaded: {uploaded_file}")
        """  # noqa: E501
        release = await self.get(repo, tag)
        asset_url = release.upload_url.replace("{?name,label}", "")

        async def upload_file(
            file_path: Path, content_type: str
        ) -> Tuple[httpx.Response, Path]:
            response = await self._client.post(
                asset_url,
                params={"name": file_path.name},
                content_type=content_type,
                content_length=file_path.stat().st_size,
                content=upload(file_path),  # type: ignore
            )
            return response, file_path

        tasks = []
        for file_path in files:
            if isinstance(file_path, tuple):
                file_path, content_type = file_path
            else:
                content_type = "application/octet-stream"

            tasks.append(upload_file(file_path, content_type))

        for coroutine in asyncio.as_completed(tasks):
            response, file_path = await coroutine
            response.raise_for_status()
            yield file_path



# Copyright (C) 2022 Greenbone AG
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# pylint: disable=redefined-builtin,disallowed-name
# ruff: noqa: E501

import unittest
from enum import Enum
from pathlib import Path
from unittest.mock import MagicMock, call, patch

import httpx

from pontos.errors import PontosError
from pontos.helper import (
    DEFAULT_TIMEOUT,
    AsyncDownloadProgressIterable,
    DownloadProgressIterable,
    add_sys_path,
    deprecated,
    download,
    download_async,
    ensure_unload_module,
    enum_or_value,
    parse_timedelta,
    snake_case,
    unload_module,
)
from pontos.testing import temp_file, temp_python_module
from tests import (
    AsyncIteratorMock,
    AsyncMock,
    IsolatedAsyncioTestCase,
    aiter,
    anext,
)
from tests.github.api import create_response


class AsyncDownloadProgressIterableTestCase(IsolatedAsyncioTestCase):
    def test_properties(self):
        content_iterator = AsyncIteratorMock(["1", "2"])

        download_iterable = AsyncDownloadProgressIterable(
            content_iterator=content_iterator, length=2, url="https://foo.bar"
        )

        self.assertEqual(download_iterable.length, 2)
        self.assertEqual(download_iterable.url, "https://foo.bar")

    async def test_download_progress(self):
        content_iterator = AsyncIteratorMock(["1", "2"])

        download_iterable = AsyncDownloadProgressIterable(
            content_iterator=content_iterator, length=2, url="https://foo.bar"
        )

        it = aiter(download_iterable)
        content, progress = await anext(it)

        self.assertEqual(content, "1")
        self.assertEqual(progress, 50)

        content, progress = await anext(it)
        self.assertEqual(content, "2")
        self.assertEqual(progress, 100)

    async def test_download_progress_without_length(self):
        content_iterator = AsyncIteratorMock(["1", "2"])

        download_iterable = AsyncDownloadProgressIterable(
            content_iterator=content_iterator,
            length=None,
            url="https://foo.bar",
        )

        it = aiter(download_iterable)
        content, progress = await anext(it)

        self.assertEqual(content, "1")
        self.assertIsNone(progress)

        content, progress = await anext(it)
        self.assertEqual(content, "2")
        self.assertIsNone(progress)


class DownloadAsyncTestCase(IsolatedAsyncioTestCase):
    async def test_download_async(self):
        response = create_response(url="http://some.url")
        response.aiter_bytes.return_value = AsyncIteratorMock(["1", "2"])
        stream = AsyncMock()
        stream.__aenter__.return_value = response

        async with download_async(
            stream, content_length=2
        ) as download_iterable:
            self.assertEqual(download_iterable.url, "http://some.url")

            it = aiter(download_iterable)
            content, progress = await anext(it)

            self.assertEqual(content, "1")
            self.assertEqual(progress, 50)

            content, progress = await anext(it)
            self.assertEqual(content, "2")
            self.assertEqual(progress, 100)

    async def test_download_async_content_length(self):
        response = create_response(headers=MagicMock(), url="http://some.url")
        response.aiter_bytes.return_value = AsyncIteratorMock(["1", "2"])
        response.headers.get.return_value = 2
        stream = AsyncMock()
        stream.__aenter__.return_value = response

        async with download_async(stream) as download_iterable:
            self.assertEqual(download_iterable.url, "http://some.url")

            it = aiter(download_iterable)
            content, progress = await anext(it)

            self.assertEqual(content, "1")
            self.assertEqual(progress, 50)

            content, progress = await anext(it)
            self.assertEqual(content, "2")
            self.assertEqual(progress, 100)

    async def test_download_async_url(self):
        response = create_response(url="http://some.url")
        response.aiter_bytes.return_value = AsyncIteratorMock(["1", "2"])
        stream = AsyncMock()
        stream.__aenter__.return_value = response

        async with download_async(
            stream,
            url="http://foo.bar",
            content_length=2,
        ) as download_iterable:
            self.assertEqual(download_iterable.url, "http://foo.bar")

            it = aiter(download_iterable)
            content, progress = await anext(it)

            self.assertEqual(content, "1")
            self.assertEqual(progress, 50)

            content, progress = await anext(it)
            self.assertEqual(content, "2")
            self.assertEqual(progress, 100)

    async def test_download_async_failure(self):
        response = create_response()
        response.raise_for_status.side_effect = httpx.HTTPStatusError(
            "404", request=MagicMock(), response=response
        )
        stream = AsyncMock()
        stream.__aenter__.return_value = response

        with self.assertRaises(httpx.HTTPStatusError):
            async with download_async(stream, content_length=2):
                pass


class DownloadProgressIterableTestCase(unittest.TestCase):
    def test_properties(self):
        content = ["foo", "bar", "baz"]
        destination = Path("foo")
        length = 123
        url = "bar"
        download_progress = DownloadProgressIterable(
            content_iterator=content,
            destination=destination,
            length=length,
            url=url,
        )

        self.assertEqual(download_progress.url, url)
        self.assertEqual(download_progress.length, length)
        self.assertEqual(download_progress.destination, destination)

    def test_progress_without_length(self):
        content = ["foo", "bar", "baz"]
        destination = MagicMock()
        writer = MagicMock()
        context_manager = MagicMock()
        context_manager.__enter__.return_value = writer
        destination.open.return_value = context_manager
        download_progress = DownloadProgressIterable(
            content_iterator=content,
            destination=destination,
            length=None,
            url="foo",
        )

        self.assertEqual(download_progress.url, "foo")

        it = iter(download_progress)
        progress = next(it)
        writer.write.assert_called_with("foo")
        self.assertIsNone(progress)
        progress = next(it)
        writer.write.assert_called_with("bar")
        self.assertIsNone(progress)
        progress = next(it)
        writer.write.assert_called_with("baz")
        self.assertIsNone(progress)

        with self.assertRaises(StopIteration):
            next(it)

    def test_progress(self):
        content = ["foo", "bar", "baz"]
        destination = MagicMock()
        writer = MagicMock()
        context_manager = MagicMock()
        context_manager.__enter__.return_value = writer
        destination.open.return_value = context_manager
        download_progress = DownloadProgressIterable(
            content_iterator=content,
            destination=destination,
            length=9,
            url="foo",
        )

        it = iter(download_progress)
        progress = next(it)
        writer.write.assert_called_with("foo")
        self.assertEqual(progress, 1 / 3.0)
        progress = next(it)
        writer.write.assert_called_with("bar")
        self.assertEqual(progress, 2 / 3.0)
        progress = next(it)
        writer.write.assert_called_with("baz")
        self.assertEqual(progress, 1)

        with self.assertRaises(StopIteration):
            next(it)

    def test_run(self):
        content = ["foo", "bar", "baz"]
        destination = MagicMock()
        writer = MagicMock()
        context_manager = MagicMock()
        context_manager.__enter__.return_value = writer
        destination.open.return_value = context_manager
        download_progress = DownloadProgressIterable(
            content_iterator=content,
            destination=destination,
            length=9,
            url="foo",
        )

        download_progress.run()
        destination.open.assert_called_once()
        writer.write.assert_has_calls((call("foo"), call("bar"), call("baz")))


class DownloadTestCase(unittest.TestCase):
    @patch("pontos.github.api.api.httpx.stream")
    def test_download_without_destination(
        self,
        requests_mock: MagicMock,
    ):
        response = MagicMock()
        response.iter_bytes.return_value = [b"foo", b"bar", b"baz"]
        response_headers = MagicMock()
        response.headers = response_headers
        response_headers.get.return_value = None
        response_stream = MagicMock()
        response_stream.__enter__.return_value = response
        requests_mock.return_value = response_stream

        with download(
            "https://github.com/greenbone/pontos/archive/refs/tags/v21.11.0.tar.gz"  # pylint: disable=line-too-long
        ) as download_progress:
            requests_mock.assert_called_once_with(
                "GET",
                "https://github.com/greenbone/pontos/archive/refs/tags/v21.11.0.tar.gz",  # pylint: disable=line-too-long
                follow_redirects=True,
                timeout=DEFAULT_TIMEOUT,
                headers=None,
                params=None,
            )
            response_headers.get.assert_called_once_with("content-length")

            self.assertIsNone(download_progress.length)
            self.assertEqual(
                download_progress.destination, Path("v21.11.0.tar.gz")
            )

            it = iter(download_progress)
            progress = next(it)
            self.assertIsNone(progress)
            progress = next(it)
            self.assertIsNone(progress)
            progress = next(it)
            self.assertIsNone(progress)

            with self.assertRaises(StopIteration):
                next(it)

            download_progress.destination.unlink()

    @patch("pontos.helper.Path")
    @patch("pontos.github.api.api.httpx.stream")
    def test_download_with_content_length(
        self, requests_mock: MagicMock, path_mock: MagicMock
    ):
        response = MagicMock()
        response.iter_bytes.return_value = [b"foo", b"bar", b"baz"]
        response_headers = MagicMock()
        response.headers = response_headers
        response_headers.get.return_value = "9"
        response_stream = MagicMock()
        response_stream.__enter__.return_value = response
        requests_mock.return_value = response_stream

        download_file = path_mock()
        file_mock = MagicMock()
        file_mock.__enter__.return_value = file_mock
        download_file.open.return_value = file_mock

        with download(
            "https://github.com/greenbone/pontos/archive/refs/tags/v21.11.0.tar.gz",  # pylint: disable=line-too-long
            download_file,
        ) as download_progress:
            requests_mock.assert_called_once_with(
                "GET",
                "https://github.com/greenbone/pontos/archive/refs/tags/v21.11.0.tar.gz",  # pylint: disable=line-too-long
                timeout=DEFAULT_TIMEOUT,
                follow_redirects=True,
                headers=None,
                params=None,
            )
            response_headers.get.assert_called_once_with("content-length")

            self.assertEqual(download_progress.length, 9)

            it = iter(download_progress)

            progress = next(it)
            self.assertEqual(progress, 1 / 3)
            file_mock.write.assert_called_with(b"foo")

            progress = next(it)
            self.assertEqual(progress, 2 / 3)
            file_mock.write.assert_called_with(b"bar")

            progress = next(it)
            self.assertEqual(progress, 1)
            file_mock.write.assert_called_with(b"baz")

            with self.assertRaises(StopIteration):
                next(it)


class DeprecatedTestCase(unittest.TestCase):
    def test_function(self):
        @deprecated
        def foo():
            pass

        with self.assertWarnsRegex(DeprecationWarning, "foo is deprecated"):
            foo()

        @deprecated()
        def foo2():
            pass

        with self.assertWarnsRegex(DeprecationWarning, "foo2 is deprecated"):
            foo2()

    def test_function_with_since(self):
        @deprecated(since="1.2.3")
        def foo():
            pass

        with self.assertWarnsRegex(
            DeprecationWarning, "deprecated since version 1.2.3"
        ):
            foo()

    def test_function_with_reason(self):
        @deprecated("Because it is obsolete.")
        def foo():
            pass

        with self.assertWarnsRegex(
            DeprecationWarning, "Because it is obsolete"
        ):
            foo()

        @deprecated(reason="Because it is obsolete.")
        def foo2():
            pass

        with self.assertWarnsRegex(
            DeprecationWarning, "Because it is obsolete"
        ):
            foo2()

    def test_class(self):
        @deprecated
        class Foo:
            pass

        with self.assertWarnsRegex(DeprecationWarning, "Foo is deprecated"):
            Foo()

        @deprecated()
        class Foo2:
            pass

        with self.assertWarnsRegex(DeprecationWarning, "Foo2 is deprecated"):
            Foo2()

    def test_class_with_since(self):
        @deprecated(since="1.2.3")
        class Foo:
            pass

        with self.assertWarnsRegex(
            DeprecationWarning, "deprecated since version 1.2.3"
        ):
            Foo()

    def test_class_with_reason(self):
        @deprecated("Because it is obsolete.")
        class Foo:
            pass

        with self.assertWarnsRegex(
            DeprecationWarning, "Because it is obsolete"
        ):
            Foo()

        @deprecated(reason="Because it is obsolete.")
        class Foo2:
            pass

        with self.assertWarnsRegex(
            DeprecationWarning, "Because it is obsolete"
        ):
            Foo2()

    def test_method(self):
        class Foo:
            @deprecated
            def bar(self):
                pass

        with self.assertWarnsRegex(DeprecationWarning, "bar is deprecated"):
            Foo().bar()

        class Foo2:
            @deprecated
            def bar(self):
                pass

        with self.assertWarnsRegex(DeprecationWarning, "bar is deprecated"):
            Foo2().bar()

    def test_method_with_since(self):
        class Foo:
            @deprecated(since="1.2.3")
            def bar(self):
                pass

        with self.assertWarnsRegex(
            DeprecationWarning, "deprecated since version 1.2.3"
        ):
            Foo().bar()

    def test_method_with_reason(self):
        class Foo:
            @deprecated("Because it is obsolete.")
            def bar(self):
                pass

        with self.assertWarnsRegex(
            DeprecationWarning, "Because it is obsolete"
        ):
            Foo().bar()

        class Foo2:
            @deprecated(reason="Because it is obsolete.")
            def bar(self):
                pass

        with self.assertWarnsRegex(
            DeprecationWarning, "Because it is obsolete"
        ):
            Foo2().bar()


class AddSysPathTestCase(unittest.TestCase):
    def test_add_sys_path(self):
        with self.assertRaises(ImportError):
            # pylint: disable=import-error,import-outside-toplevel,unused-import
            import mymodule  # noqa: F811,F401

        with temp_file("", name="mymodule.py") as module_path, add_sys_path(
            module_path.parent
        ):
            # pylint: disable=import-error,import-outside-toplevel,unused-import
            import mymodule  # noqa: F811,F401

        unload_module("mymodule")


class EnsureUnloadModuleTestCase(unittest.TestCase):
    def test_ensure_unload_module(self):
        with temp_python_module(
            "def foo():\n  pass", name="bar"
        ), ensure_unload_module("bar"):
            # pylint: disable=import-error,import-outside-toplevel,unused-import
            import bar  # noqa: F401,F811

        with self.assertRaises(ImportError):
            # pylint: disable=import-error,import-outside-toplevel,unused-import
            import bar  # noqa: F401,F811

    def test_ensure_unload_module_exception(self):
        with self.assertRaisesRegex(ValueError, "Ipsum"):
            with temp_python_module(
                "def func():\n  raise ValueError('Ipsum')", name="bar"
            ), ensure_unload_module("bar"):
                # pylint: disable=import-error,import-outside-toplevel,unused-import
                import bar

                bar.func()

        with self.assertRaises(ImportError):
            # pylint: disable=import-error,import-outside-toplevel,unused-import
            import bar

    def test_add_sys_path_exception(self):
        with self.assertRaises(ImportError):
            # pylint: disable=import-error,import-outside-toplevel,unused-import
            import mymodule  # noqa: F811,F401

        try:
            with temp_file("", name="mymodule.py") as module_path, add_sys_path(
                module_path.parent
            ):
                # pylint: disable=import-error,import-outside-toplevel,unused-import
                import mymodule  # noqa: F811,F401

                raise ValueError()
        except ValueError:
            pass
        finally:
            unload_module("mymodule")

        with self.assertRaises(ImportError):
            # pylint: disable=import-error,import-outside-toplevel,unused-import
            import mymodule  # noqa: F401,F811


class SnakeCaseTestCase(unittest.TestCase):
    def test_snake_case(self):
        self.assertEqual(snake_case("CamelCase"), "camel_case")
        self.assertEqual(snake_case("camelCase"), "camel_case")
        self.assertEqual(snake_case("snakecase"), "snakecase")
        self.assertEqual(snake_case("snake_case"), "snake_case")


class EnumOrValueTestCase(unittest.TestCase):
    def test_value(self):
        self.assertEqual(enum_or_value(None), None)
        self.assertEqual(enum_or_value("foo"), "foo")
        self.assertEqual(enum_or_value(123), 123)

    def test_enum(self):
        class Foo(Enum):
            BAR = "bar"
            BAZ = "baz"

        self.assertEqual(enum_or_value(Foo.BAR), "bar")
        self.assertEqual(enum_or_value(Foo.BAZ), "baz")


class ParseTimedeltaTestCase(unittest.TestCase):
    def test_parse_complex(self):
        delta = parse_timedelta("1w2d4h5m6s")

        self.assertEqual(delta.days, 9)
        self.assertEqual(delta.seconds, 14706)

    def test_parse_weeks(self):
        delta = parse_timedelta("1w")
        self.assertEqual(delta.days, 7)
        self.assertEqual(delta.seconds, 0)

        delta = parse_timedelta("1.5w")
        self.assertEqual(delta.days, 10)
        self.assertEqual(delta.seconds, 43200)

    def test_parse_days(self):
        delta = parse_timedelta("1d")
        self.assertEqual(delta.days, 1)
        self.assertEqual(delta.seconds, 0)

        delta = parse_timedelta("1.5d")
        self.assertEqual(delta.days, 1)
        self.assertEqual(delta.seconds, 43200)

    def test_parse_hours(self):
        delta = parse_timedelta("1h")
        self.assertEqual(delta.days, 0)
        self.assertEqual(delta.seconds, 3600)

        delta = parse_timedelta("1.5h")
        self.assertEqual(delta.days, 0)
        self.assertEqual(delta.seconds, 5400)

    def test_parse_error(self):
        with self.assertRaises(PontosError):
            parse_timedelta("foo")

        with self.assertRaises(PontosError):
            parse_timedelta("1d2x")

        with self.assertRaises(PontosError):
            parse_timedelta("1,2d")
